from typing import Type, Optional
from pydantic.dataclasses import dataclass
from iterwrapper import IterWrapper
from pydantic.main import BaseModel, ModelMetaclass

import copy


class EventMeta(ModelMetaclass):
    def __new__(mcls, name, bases, mapping, **kwargs):
        if any(
            IterWrapper(bases)
            .filter(lambda x: getattr(x, "__base_event__", False))
            .collect(list)
        ):
            if not mapping.__contains__("__base_event__"):
                mapping["__base_event__"] = False
        if not mapping.get("Dispatcher") and name != "BaseEvent":
            raise AttributeError(
                "a event class must have a dispatcher called 'Dispatcher'"
            )
        r = super().__new__(
            mcls,
            name,
            (BaseModel, *bases) if name == "BaseEvent" else bases,
            mapping,
            **kwargs
        )
        if mapping.get("type"):
            r.type = mapping.get("type")
        if name != "BaseEvent":
            r.update_forward_refs()
            r.Dispatcher = mapping["Dispatcher"]
        return r


class BaseEvent(metaclass=EventMeta):
    class Config:
        arbitrary_types_allowed = True


from .dispatcher import BaseDispatcher



from abc import ABCMeta, abstractmethod, abstractproperty, abstractstaticmethod
from functools import lru_cache
from inspect import ismethod
import traceback
from typing import (
    Any,
    AsyncGenerator,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

from graia.broadcast.utilles import run_always_await_safely, cached_getattr

DEFAULT_LIFECYCLE_NAMES = (
    "beforeDispatch",
    "afterDispatch",
    "beforeExecution",
    "afterExecution",
    "beforeTargetExec",
    "afterTargetExec",
)


class IDispatcherInterface(metaclass=ABCMeta):
    broadcast: "Broadcast"

    execution_contexts: List["ExecutionContext"]
    parameter_contexts: List["ParameterContext"]
    alive_generator_dispatcher: List[
        List[Tuple[Union[Generator, AsyncGenerator], bool]]
    ]

    @abstractmethod
    def start_execution(
        self,
        event: "BaseEvent",
        dispatchers: List["T_Dispatcher"],
        use_inline_generator: bool = False,
    ):
        pass

    @abstractmethod
    async def exit_current_execution(self):
        pass

    @abstractmethod
    async def alive_dispatcher_killer(self):
        pass

    def inject_local_raw(self, *dispatchers: List["T_Dispatcher"], source: Any = None):
        # 为什么没有 flush: 因为这里的 lifecycle 是无意义的.
        for dispatcher in dispatchers[::-1]:
            self.parameter_contexts[-1].dispatchers.insert(0, dispatcher)
        always_dispatchers = self.execution_contexts[-1].always_dispatchers

        for i in dispatchers:
            if cached_getattr(i, "always", False):
                always_dispatchers.add(i)

    def inject_execution_raw(self, *dispatchers: List["T_Dispatcher"]):
        for dispatcher in dispatchers:
            self.execution_contexts[-1].dispatchers.insert(0, dispatcher)
        always_dispatchers = self.execution_contexts[-1].always_dispatchers

        self.flush_lifecycle_refs(dispatchers)
        for i in dispatchers:
            if cached_getattr(i, "always", False):
                always_dispatchers.add(i)

    def inject_global_raw(self, *dispatchers: List["T_Dispatcher"]):
        # self.dispatchers.extend(dispatchers)
        for dispatcher in dispatchers[::-1]:
            self.execution_contexts[0].dispatchers.insert(1, dispatcher)
        always_dispatchers = self.execution_contexts[-1].always_dispatchers

        self.flush_lifecycle_refs(dispatchers)
        for i in dispatchers:
            if cached_getattr(i, "always", False):
                always_dispatchers.add(i)

    @staticmethod
    @lru_cache
    def get_lifecycle_refs(
        dispatcher: "T_Dispatcher",
    ) -> Optional[Dict[str, List]]:
        from graia.broadcast.entities.dispatcher import BaseDispatcher

        lifecycle_refs: Dict[str, List] = {}
        if not isinstance(dispatcher, (BaseDispatcher, type)):
            return

        for name in DEFAULT_LIFECYCLE_NAMES:
            lifecycle_refs.setdefault(name, [])
            abstract_lifecycle_func = cached_getattr(BaseDispatcher, name)
            unbound_attr = getattr(dispatcher, name, None)

            if unbound_attr is None:
                continue

            orig_call = unbound_attr
            while ismethod(orig_call):
                orig_call = unbound_attr.__func__

            if orig_call is abstract_lifecycle_func:
                continue

            lifecycle_refs[name].append(unbound_attr)

        return lifecycle_refs

    def flush_lifecycle_refs(
        self,
        dispatchers: List["T_Dispatcher"] = None,
    ):
        from graia.broadcast.entities.dispatcher import BaseDispatcher

        lifecycle_refs = self.execution_contexts[-1].lifecycle_refs
        if dispatchers is None and lifecycle_refs:  # 已经刷新.
            return

        for dispatcher in dispatchers or self.dispatcher_pure_generator():
            if (
                not isinstance(dispatcher, BaseDispatcher)
                or dispatcher.__class__ is type
            ):
                continue

            for name, value in self.get_lifecycle_refs(dispatcher).items():
                lifecycle_refs.setdefault(name, [])
                lifecycle_refs[name].extend(value)

    async def exec_lifecycle(self, lifecycle_name: str, *args, **kwargs):
        lifecycle_funcs = self.execution_contexts[-1].lifecycle_refs.get(
            lifecycle_name, []
        )
        if lifecycle_funcs:
            for func in lifecycle_funcs:
                await run_always_await_safely(func, self, *args, **kwargs)

    @abstractproperty
    def dispatcher_sources(self) -> List["DispatcherSource"]:
        pass

    @abstractmethod
    def dispatcher_pure_generator(self) -> Generator[None, None, "T_Dispatcher"]:
        pass

    @abstractmethod
    def dispatcher_generator(
        self,
    ) -> Generator[None, None, Tuple["T_Dispatcher", "T_Dispatcher_Callable", Any]]:
        pass

    @abstractproperty
    def name(self) -> str:
        pass

    @abstractproperty
    def annotation(self) -> Any:
        pass

    @abstractproperty
    def default(self) -> Any:
        pass

    @abstractproperty
    def _index(self) -> int:
        pass

    @abstractproperty
    def global_dispatcher(self) -> List["T_Dispatcher"]:
        pass

    @abstractproperty
    def event(self) -> "BaseEvent":
        pass

    @abstractmethod
    async def execute_dispatcher_callable(
        self, dispatcher_callable: "T_Dispatcher_Callable"
    ) -> Any:
        pass

    @abstractmethod
    async def lookup_param(self, name: str, annotation: Any, default: Any) -> Any:
        pass

    @abstractmethod
    async def lookup_using_current(self) -> Any:
        pass

    @abstractmethod
    async def lookup_by(
        self, dispatcher: "T_Dispatcher", name: str, annotation: Any, default: Any
    ) -> Any:
        pass

    @abstractmethod
    async def lookup_by_directly(
        self, dispatcher: "T_Dispatcher", name: str, annotation: Any, default: Any
    ) -> Any:
        pass

    @abstractproperty
    def has_current_exec_context(self) -> bool:
        pass

    @abstractproperty
    def has_current_param_context(self) -> bool:
        pass


from graia.broadcast.typing import T_Dispatcher, T_Dispatcher_Callable
from graia.broadcast.entities.context import ExecutionContext, ParameterContext
from graia.broadcast.entities.source import DispatcherSource
from graia.broadcast.entities.event import BaseEvent



from abc import ABCMeta, abstractstaticmethod
from types import TracebackType
from typing import List, Optional


class BaseDispatcher(metaclass=ABCMeta):
    always: bool
    mixin: List["BaseDispatcher"]

    @abstractstaticmethod
    def catch(interface: "IDispatcherInterface"):
        pass

    def beforeDispatch(self, interface: "IDispatcherInterface"):
        pass

    def afterDispatch(self, interface: "IDispatcherInterface"):
        pass

    def beforeExecution(self, interface: "IDispatcherInterface"):
        pass

    def afterExecution(
        self,
        interface: "IDispatcherInterface",
        exception: Optional[Exception],
        tb: Optional[TracebackType],
    ):
        pass

    def beforeTargetExec(self, interface: "IDispatcherInterface"):
        pass

    def afterTargetExec(
        self,
        interface: "IDispatcherInterface",
        exception: Optional[Exception],
        tb: Optional[TracebackType],
    ):
        pass

    def onActive(self, interface: "IDispatcherInterface"):
        pass


from graia.broadcast.abstract.interfaces.dispatcher import IDispatcherInterface



import weakref

from typing import Any, Callable, Dict, List, Set, Type, Union
from graia.broadcast.entities.source import DispatcherSource
from graia.broadcast.utilles import cached_isinstance

from ..typing import T_Dispatcher


class ExecutionContext:
    source: DispatcherSource[T_Dispatcher, "ExecutionContext"]
    always_dispatchers: Set[Union["BaseDispatcher", Type["BaseDispatcher"]]]
    event: "BaseEvent"
    inline_generator: bool
    _index: int
    lifecycle_refs: Dict[str, List[Callable]]

    def __init__(
        self,
        dispatchers: List[T_Dispatcher],
        event: "BaseEvent",
        inline_generator: bool = False,
    ) -> None:
        self.source = DispatcherSource(dispatchers, weakref.ref(self))
        self.event = event
        self.inline_generator = inline_generator
        self._index = 0

        self.lifecycle_refs = {}
        self.always_dispatchers = set()
        self.always_dispatchers.update(
            filter(
                lambda x: cached_isinstance(x, BaseDispatcher) and x.always, dispatchers
            )
        )

    @property
    def dispatchers(self):
        return self.source.dispatchers


class ParameterContext:
    def __init__(self, name, annotation, default, dispatchers) -> None:
        self.name, self.annotation, self.default, self.source = (
            name,
            annotation,
            default,
            DispatcherSource(dispatchers, weakref.ref(self)),
        )

    def __repr__(self) -> str:
        return (
            "<ParameterContext name={0} annotation={1} default={2} locald={3}".format(
                self.name, self.annotation, self.default, self.dispatchers
            )
        )

    source: DispatcherSource[T_Dispatcher, "ParameterContext"]

    name: str
    annotation: Any
    default: Any

    @property
    def dispatchers(self) -> List[T_Dispatcher]:
        return self.source.dispatchers


from graia.broadcast.entities.dispatcher import BaseDispatcher
from .event import BaseEvent



from typing import Any, Dict

from graia.broadcast.entities.dispatcher import BaseDispatcher

from ..entities.decorater import Decorater
from ..entities.signatures import Force
from ..utilles import (
    is_asyncgener,
    iscoroutinefunction,
    isgeneratorfunction,
    run_always_await_safely,
    cached_isinstance,
)
from ..abstract.interfaces.dispatcher import IDispatcherInterface


class DecoraterInterface(BaseDispatcher):
    """Graia Broadcast Control 内部机制 Decorate 的具体管理实现"""

    dispatcher_interface: IDispatcherInterface
    local_storage: Dict[Any, Any] = {}
    return_value: Any = None
    default = None

    def __init__(self, dispatcher_interface: IDispatcherInterface):
        self.dispatcher_interface = dispatcher_interface

    @property
    def name(self):
        return self.dispatcher_interface.name

    @property
    def annotation(self):
        return self.dispatcher_interface.annotation

    @property
    def event(self):
        return self.dispatcher_interface.event

    async def catch(self, interface: IDispatcherInterface):
        if cached_isinstance(interface.default, Decorater):
            decorater: Decorater = interface.default
            if not decorater.pre:
                # 作为 装饰
                self.return_value = await interface.lookup_param(
                    interface.name, interface.annotation, None
                )
            try:
                # 这里隐式的复用了 dispatcher interface 的生成器终结者机制
                if is_asyncgener(decorater.target):
                    # 如果是异步生成器
                    async for i in decorater.target(self):
                        yield i
                elif isgeneratorfunction(decorater.target) and not iscoroutinefunction(
                    decorater.target
                ):
                    # 同步生成器
                    for i in decorater.target(self):
                        yield i
                else:
                    yield Force(await run_always_await_safely(decorater.target, self))
            finally:
                if not decorater.pre:
                    self.return_value = None



from typing import Any, Callable, Type, Union

T_Dispatcher = Union[
    Type["BaseDispatcher"], "BaseDispatcher", Callable[["IDispatcherInterface"], Any]
]

T_Dispatcher_Callable = Callable[["IDispatcherInterface"], Any]

from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.abstract.interfaces.dispatcher import IDispatcherInterface



from typing import Any, Callable


class Decorater:
    target: Callable[[Any], Any]
    pre: bool = False

    def __init__(self, target, pre=False):
        super().__init__(target=target, pre=pre)



import traceback
import weakref
import itertools
from functools import lru_cache

from graia.broadcast.abstract.interfaces.dispatcher import IDispatcherInterface
from typing import (
    Any,
    Callable,
    Generator,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)
from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.entities.event import BaseEvent

from graia.broadcast.entities.context import ExecutionContext, ParameterContext
from graia.broadcast.entities.signatures import Force
from graia.broadcast.entities.source import DispatcherSource
from graia.broadcast.exceptions import OutOfMaxGenerater, RequirementCrashed
from graia.broadcast.typing import T_Dispatcher, T_Dispatcher_Callable
from graia.broadcast.utilles import (
    is_asyncgener,
    isgeneratorfunction,
    run_always_await_safely,
    cached_getattr,
)


class EmptyEvent(BaseEvent):
    class Dispatcher(BaseDispatcher):
        @staticmethod
        def catch(_):
            pass


class DispatcherInterface(IDispatcherInterface):
    @property
    def name(self) -> str:
        return self.parameter_contexts[-1].name

    @property
    def annotation(self) -> Any:
        return self.parameter_contexts[-1].annotation

    @property
    def default(self) -> Any:
        return self.parameter_contexts[-1].default

    @property
    def _index(self) -> int:
        return self.execution_contexts[-1]._index

    @property
    def event(self) -> BaseEvent:
        return self.execution_contexts[-1].event

    @property
    def global_dispatcher(self) -> List[T_Dispatcher]:
        return self.execution_contexts[0].dispatchers

    @property
    def has_current_exec_context(self) -> bool:
        return len(self.execution_contexts) >= 2

    @property
    def has_current_param_context(self) -> bool:
        return len(self.parameter_contexts) >= 2

    def __init__(self, broadcast_instance: "Broadcast") -> None:
        self.broadcast = broadcast_instance
        self.alive_generator_dispatcher = [[]]
        self.execution_contexts = [ExecutionContext([], EmptyEvent())]
        self.parameter_contexts = [ParameterContext(None, None, None, [])]

    async def __aenter__(self) -> "DispatcherInterface":
        return self

    async def __aexit__(self, _, exc: Exception, tb):
        await self.exit_current_execution()
        if tb is not None:
            raise exc.with_traceback(tb)

    def start_execution(
        self,
        event: BaseEvent,
        dispatchers: List[T_Dispatcher],
        use_inline_generator: bool = False,
    ) -> "DispatcherInterface":
        self.execution_contexts.append(
            ExecutionContext(dispatchers, event, use_inline_generator)
        )
        self.alive_generator_dispatcher.append([])
        self.flush_lifecycle_refs()
        return self

    async def exit_current_execution(self):
        current_generators = self.alive_generator_dispatcher[-1]
        if current_generators:
            if self.execution_contexts[-1].inline_generator:
                if len(self.alive_generator_dispatcher) > 1:  # 防止插入到保护区
                    self.alive_generator_dispatcher[-2].extend(current_generators)
                else:
                    raise ValueError("cannot cast to inline")
            else:
                await self.alive_dispatcher_killer()
        self.alive_generator_dispatcher.pop()
        self.execution_contexts.pop()

    async def alive_dispatcher_killer(self):
        for unbound_gen, is_async_gen in self.alive_generator_dispatcher[-1]:
            index = 0
            if is_async_gen:
                async for _ in unbound_gen:
                    if index == 15:
                        raise OutOfMaxGenerater(
                            "a dispatch as a sync generator had to stop: ", unbound_gen
                        )
                    index += 1
            else:
                for _ in unbound_gen:
                    if index == 15:
                        raise OutOfMaxGenerater(
                            "a dispatch as a sync generator had to stop: ", unbound_gen
                        )
                    index += 1

    @property
    def dispatcher_sources(self) -> List[DispatcherSource]:
        return [
            self.execution_contexts[0].source,
            self.parameter_contexts[-1].source,
            self.execution_contexts[-1].source,
        ]

    def dispatcher_pure_generator(self) -> Generator[None, None, T_Dispatcher]:
        for source in self.dispatcher_sources:
            yield from source.dispatchers

    @staticmethod
    @lru_cache(None)
    def dispatcher_callable_detector(dispatcher: T_Dispatcher) -> T_Dispatcher_Callable:
        if dispatcher.__class__ is type and issubclass(dispatcher, BaseDispatcher):
            return dispatcher().catch
        elif hasattr(dispatcher, "catch"):
            return dispatcher.catch
        elif callable(dispatcher):
            return dispatcher
        else:
            raise ValueError("invaild dispatcher: ", dispatcher)

    def dispatcher_generator(
        self,
        source_from: Iterator[DispatcherSource[T_Dispatcher, Any]] = None,
        using_always: bool = True,
    ) -> Generator[
        None, None, Tuple[T_Dispatcher, T_Dispatcher_Callable, DispatcherSource]
    ]:
        always_dispatcher = self.execution_contexts[-1].always_dispatchers
        if source_from is None:
            source_from = self.dispatcher_sources

        for source in source_from:
            for dispatcher in source.dispatchers:
                if using_always and dispatcher in always_dispatcher:
                    always_dispatcher.remove(dispatcher)

                dispatcher_callable = self.dispatcher_callable_detector(dispatcher)
                yield (dispatcher, dispatcher_callable, source)

    async def execute_dispatcher_callable(
        self, dispatcher_callable: T_Dispatcher_Callable
    ) -> Any:
        alive_dispatchers = self.alive_generator_dispatcher[-1]

        if is_asyncgener(dispatcher_callable):
            current_generator = dispatcher_callable(self).__aiter__()

            try:
                result = await current_generator.__anext__()
            except StopAsyncIteration:
                return
            else:
                alive_dispatchers.append((current_generator, True))
        elif isgeneratorfunction(dispatcher_callable):
            current_generator = dispatcher_callable(self).__iter__()

            try:
                result = current_generator.__next__()
            except StopIteration as e:
                result = e.value
            else:
                alive_dispatchers.append((current_generator, False))
        else:
            result = await run_always_await_safely(dispatcher_callable, self)

        return result

    async def lookup_param(
        self,
        name: str,
        annotation: Any,
        default: Any,
        enable_extra_return: bool = False,
    ) -> Union[Any, Tuple[Any, T_Dispatcher, DispatcherSource, List[T_Dispatcher]]]:
        self.parameter_contexts.append(ParameterContext(name, annotation, default, []))

        result = None
        try:
            start_offset = self._index + int(bool(self._index))
            for self.execution_contexts[-1]._index, (
                dispatcher,
                dispatcher_callable,
                source,
            ) in enumerate(
                itertools.islice(self.dispatcher_generator(), start_offset, None, None),
                start=start_offset,
            ):
                result = await self.execute_dispatcher_callable(dispatcher_callable)

                if result is None:
                    continue

                if result.__class__ is Force:
                    result = result.target

                self.execution_contexts[-1]._index = 0
                return (
                    (
                        result,
                        weakref.ref(dispatcher),
                        source,
                        list(
                            map(
                                lambda x: weakref.ref(x),
                                itertools.islice(
                                    self.dispatcher_generator(),
                                    start_offset,
                                    self._index,
                                ),
                            )
                        ),
                    )
                    if enable_extra_return
                    else result
                )
            else:
                raise RequirementCrashed(
                    "the dispatching requirement crashed: ",
                    self.name,
                    self.annotation,
                    self.default,
                )
        finally:
            always_dispatchers = self.execution_contexts[-1].always_dispatchers
            if always_dispatchers:
                for (
                    dispatcher,
                    dispatcher_callable,
                    source,
                ) in self.dispatcher_generator(
                    [DispatcherSource(always_dispatchers)],
                    using_always=False,
                ):
                    await self.execute_dispatcher_callable(dispatcher_callable)

            self.parameter_contexts.pop()

    async def lookup_using_current(self) -> Any:
        result = None
        try:
            for self.execution_contexts[-1]._index, (
                dispatcher,
                dispatcher_callable,
                source,
            ) in enumerate(
                itertools.islice(
                    self.dispatcher_generator(),
                    self._index + int(bool(self._index)),
                    None,
                    None,
                ),
                start=self._index + int(bool(self._index)),
            ):
                result = await self.execute_dispatcher_callable(dispatcher_callable)

                if result is None:
                    continue

                if result.__class__ is Force:
                    result = result.target

                self.execution_contexts[-1]._index = 0
                return result
            else:
                raise RequirementCrashed(
                    "the dispatching requirement crashed: ",
                    self.name,
                    self.annotation,
                    self.default,
                )
        finally:
            for dispatcher, dispatcher_callable, source in self.dispatcher_generator(
                [DispatcherSource(self.execution_contexts[-1].always_dispatchers)],
                using_always=False,
            ):
                await self.execute_dispatcher_callable(dispatcher_callable)

    async def lookup_by(
        self,
        dispatcher: T_Dispatcher,
        name: str,
        annotation: Any,
        default: Any,
        enable_extra_return: bool = False,
    ) -> Union[Any, Tuple[Any, T_Dispatcher, DispatcherSource, List[T_Dispatcher]]]:
        self.parameter_contexts.append(ParameterContext(name, annotation, default, []))

        result = None
        try:
            start_offset = self._index + int(bool(self._index))
            for self.execution_contexts[-1]._index, (
                dispatcher,
                dispatcher_callable,
                source,
            ) in enumerate(
                itertools.islice(
                    self.dispatcher_generator([DispatcherSource([dispatcher])]),
                    start_offset,
                    None,
                    None,
                ),
                start=start_offset,
            ):
                result = await self.execute_dispatcher_callable(dispatcher_callable)

                if result is None:
                    continue

                if result.__class__ is Force:
                    result = result.target

                self.execution_contexts[-1]._index = 0
                return (
                    (
                        result,
                        weakref.ref(dispatcher),
                        source,
                        list(
                            map(
                                lambda x: weakref.ref(x),
                                itertools.islice(
                                    self.dispatcher_generator(),
                                    start_offset,
                                    self._index,
                                ),
                            )
                        ),
                    )
                    if enable_extra_return
                    else result
                )
            else:
                raise RequirementCrashed(
                    "the dispatching requirement crashed: ",
                    self.name,
                    self.annotation,
                    self.default,
                )
        finally:
            for dispatcher, dispatcher_callable, source in self.dispatcher_generator(
                [DispatcherSource(self.execution_contexts[-1].always_dispatchers)],
                using_always=False,
            ):
                await self.execute_dispatcher_callable(dispatcher_callable)

            self.parameter_contexts.pop()

    async def lookup_by_directly(
        self, dispatcher: T_Dispatcher, name: str, annotation: Any, default: Any
    ) -> Any:
        self.parameter_contexts.append(ParameterContext(name, annotation, default, []))

        dispatcher_callable = self.dispatcher_callable_detector(dispatcher)

        try:
            result = await self.execute_dispatcher_callable(dispatcher_callable)
            if result.__class__ is Force:
                return result.target

            return result
        finally:
            self.parameter_contexts.pop()



import asyncio
from functools import lru_cache
import inspect
import sys
import traceback
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Hashable,
    Iterable,
    List,
    NoReturn,
    Optional,
    Tuple,
    Type,
    Union,
)

from iterwrapper import IterWrapper as iw

from .typing import T_Dispatcher, T_Dispatcher_Callable

from .entities.exectarget import ExecTarget

from .interfaces.dispatcher import DispatcherInterface
from .entities.decorater import Decorater
from .entities.dispatcher import BaseDispatcher
from .entities.event import BaseEvent
from .entities.inject_rule import BaseRule
from .entities.listener import Listener
from .entities.namespace import Namespace
from .entities.signatures import Force, RemoveMe
from .exceptions import (
    DisabledNamespace,
    ExecutionStop,
    ExistedNamespace,
    InvaildEventName,
    PropagationCancelled,
    RegisteredEventListener,
    RequirementCrashed,
    UnexistedNamespace,
)
from .interfaces.decorater import DecoraterInterface
from .utilles import (
    argument_signature,
    dispatcher_mixin_handler,
    group_dict,
    isasyncgen,
    isgenerator,
    printer,
    run_always_await_safely,
    cached_isinstance,
)
from .typing import T_Dispatcher
from .zone import Zone


class Broadcast:
    loop: asyncio.AbstractEventLoop

    default_namespace: Namespace
    namespaces: List[Namespace]
    listeners: List[Listener]

    # 规则驱动注入
    dispatcher_inject_rules: List[BaseRule] = []

    dispatcher_interface: DispatcherInterface

    debug_flag: bool

    use_dispatcher_statistics: bool
    use_reference_optimization: bool

    def __init__(
        self,
        *,
        loop: asyncio.AbstractEventLoop = None,
        debug_flag: bool = False,
        inject_rules: Optional[List[BaseRule]] = None,
        use_dispatcher_statistics: bool = False,
        use_reference_optimization: bool = False
    ):
        self.loop = loop or asyncio.get_event_loop()
        self.default_namespace = Namespace(name="default", default=True)
        self.debug_flag = debug_flag
        self.namespaces = []
        self.listeners = []
        self.dispatcher_inject_rules = inject_rules or []
        self.dispatcher_interface = DispatcherInterface(self)

        self.use_dispatcher_statistics = use_dispatcher_statistics

        if not use_dispatcher_statistics and use_reference_optimization:
            raise ValueError(
                "the feature of reference optimization requires dispatcher statictics."
            )
        self.use_reference_optimization = use_reference_optimization

        @self.dispatcher_interface.inject_global_raw
        def _(interface: DispatcherInterface):
            if interface.annotation is interface.event.__class__:
                return interface.event

    def default_listener_generator(self, event_class) -> Iterable[Listener]:
        return (
            iw(self.listeners)
            .filter(lambda x: not x.namespace.hide)  # filter for hide
            .filter(lambda x: not x.namespace.disabled)  # filter for disabled
            .filter(lambda x: event_class in x.listening_events)
            # .collect(list)  # collect to a whole list
        )

    def addInjectionRule(self, rule: BaseRule) -> NoReturn:
        if rule in self.dispatcher_inject_rules:
            raise ValueError("this rule has already been added!")
        self.dispatcher_inject_rules.append(rule)

    def removeInjectionRule(self, rule: BaseRule) -> NoReturn:
        self.dispatcher_inject_rules.remove(rule)

    async def layered_scheduler(
        self, listener_generator: Generator[Listener, None, None], event: BaseEvent
    ):
        grouped: Dict[int, List[Listener]] = group_dict(
            listener_generator, lambda x: x.priority
        )
        for _, current_group in sorted(grouped.items(), key=lambda x: x[0]):
            coros = [self.Executor(target=i, event=event) for i in current_group]
            try:
                await asyncio.gather(*coros)
            except Exception as e:
                if cached_isinstance(e, PropagationCancelled):
                    break

    async def Executor(
        self,
        target: Union[Callable, ExecTarget],
        event: BaseEvent,
        dispatchers: List[
            Union[
                Type[BaseDispatcher],
                Callable,
                BaseDispatcher,
            ]
        ] = None,
        post_exception_event: bool = False,
        print_exception: bool = True,
        enableInternalAccess: bool = False,
        use_inline_generator: bool = False,
        use_dispatcher_statistics: bool = False,
        use_reference_optimization: bool = False,
    ):
        from .builtin.event import ExceptionThrowed

        is_exectarget = cached_isinstance(target, ExecTarget)
        is_listener = cached_isinstance(target, Listener)

        use_dispatcher_statistics = (
            self.use_dispatcher_statistics or use_dispatcher_statistics
        )
        use_reference_optimization = (
            self.use_reference_optimization or use_reference_optimization
        )

        if is_listener:
            if target.namespace.disabled:
                raise DisabledNamespace(
                    "catched a disabled namespace: {0}".format(target.namespace.name)
                )

        if use_dispatcher_statistics and not is_exectarget:
            use_dispatcher_statistics = False
        if use_reference_optimization and not use_dispatcher_statistics:
            raise ValueError(
                "the feature of reference optimization requires dispatcher statictics."
            )

        # 先生成 dispatchers
        _dispatchers = dispatcher_mixin_handler(event.Dispatcher)
        # 开始暴力注入
        if is_exectarget:
            if target.inline_dispatchers:
                _dispatchers = target.inline_dispatchers + _dispatchers
            if is_listener:
                if target.namespace.injected_dispatchers:
                    _dispatchers = target.namespace.injected_dispatchers + _dispatchers
        if dispatchers:
            _dispatchers = dispatchers + _dispatchers

        target_callable = target.callable if is_exectarget else target
        parameter_compile_result = {}

        async with self.dispatcher_interface.start_execution(
            event, _dispatchers, use_inline_generator
        ) as dii:
            try:
                catched_first_dispatcher = next(dii.dispatcher_pure_generator())
            except StopIteration:
                dei = DecoraterInterface(dii)  # pylint: disable=unused-variable
                self.dispatcher_interface.execution_contexts[
                    0
                ].source.dispatchers.insert(0, dei)
            else:
                if not cached_isinstance(catched_first_dispatcher, DecoraterInterface):
                    dei = DecoraterInterface(dii)  # pylint: disable=unused-variable
                    self.dispatcher_interface.execution_contexts[
                        0
                    ].source.dispatchers.insert(0, dei)
            # Decorater 的 Dispatcher 已经注入, 没他事了

            if enableInternalAccess or (
                is_exectarget and target.enable_internal_access
            ):
                internal_access_mapping = {Broadcast: self, DispatcherInterface: dii}
                if is_exectarget:
                    internal_access_mapping[target.__class__] = target
                if is_listener:
                    internal_access_mapping[Namespace] = target.namespace

                @dii.inject_execution_raw
                def _(interface: DispatcherInterface):
                    return internal_access_mapping.get(interface.annotation)

            for injection_rule in self.dispatcher_inject_rules:
                if injection_rule.check(event, dii):
                    dii.execution_contexts[-1].dispatchers.insert(
                        1, injection_rule.target_dispatcher
                    )

            if is_exectarget:
                whole_statistics: Dict[
                    str, Dict[T_Dispatcher, Tuple[int, int]]
                ] = target.dispatcher_statistics["statistics"]
            else:
                whole_statistics = {}

            await dii.exec_lifecycle("beforeExecution")
            await dii.exec_lifecycle("beforeDispatch")
            try:
                for name, annotation, default in argument_signature(target_callable):
                    statistics: Dict[
                        T_Dispatcher, Tuple[int, int]
                    ] = whole_statistics.setdefault(name, {})
                    if use_dispatcher_statistics:
                        if (
                            use_reference_optimization
                        ):  # 启用被动性质的优化特性 引用缓存(Reference Cache)
                            for dispatcher, this_statistics in sorted(
                                statistics.items(), key=lambda x: x[1][0] / x[1][1]
                            ):
                                this_statistics[1] += 1
                                if dispatcher() is None:  # ref 在对象 dead 时返回 None.
                                    del statistics[dispatcher]
                                    continue

                                try:
                                    result = await dii.lookup_by_directly(
                                        dispatcher(), name, annotation, default
                                    )
                                except RequirementCrashed:  # 忽略单个 Dispatcher 执行时若无法满足要求时所引发的错误
                                    pass
                                except:
                                    traceback.print_exc()
                                    raise
                                else:
                                    this_statistics[0] += 1
                                    parameter_compile_result[name] = result
                            else:
                                if parameter_compile_result.get(name):
                                    continue

                        (
                            parameter_compile_result[name],
                            target_dispatcher,
                            target_source,
                            past_dispatchers,
                        ) = await dii.lookup_param(
                            name, annotation, default, enable_extra_return=True
                        )

                        # TODO: 判定 target_source 来源的可靠性(毕竟...)
                        if target_source.source() is not None:
                            # 1.成功给出结果的次数, 2.总共被 call 的次数
                            statistics.setdefault(target_dispatcher, [0, 0])
                            this_statistics = statistics[target_dispatcher]
                            this_statistics[0] += 1
                            this_statistics[1] += 1

                            if len(statistics) > 1:
                                past_set = set(statistics.keys())
                                for past_dispatcher in past_dispatchers:
                                    if past_dispatcher in past_set:
                                        statistics[past_dispatcher][1] += 1
                    else:
                        parameter_compile_result[name] = await dii.lookup_param(
                            name, annotation, default
                        )
                if is_exectarget:
                    if target.headless_decoraters:
                        for hl_d in target.headless_decoraters:
                            await dii.lookup_param(None, None, hl_d)
            except ExecutionStop:
                raise
            except RequirementCrashed:
                traceback.print_exc()
                raise
            except Exception as e:
                if print_exception:
                    traceback.print_exc()
                if post_exception_event:
                    self.postEvent(ExceptionThrowed(exception=e, event=event))
                raise
            finally:
                await dii.exec_lifecycle("afterDispatch")

            await dii.exec_lifecycle("beforeTargetExec")

            try:
                result = await run_always_await_safely(
                    target_callable, **parameter_compile_result
                )
            except ExecutionStop:
                raise  # 直接抛出.
            except PropagationCancelled:
                raise  # 防止事件广播
            except Exception as e:
                if print_exception:
                    traceback.print_exc()
                if post_exception_event:  # 如果没有referrer, 则广播事件, 如果有则向上抛出(防止重复抛出事件)
                    self.postEvent(ExceptionThrowed(exception=e, event=event))
                raise
            finally:
                _, exception, tb = sys.exc_info()
                await dii.exec_lifecycle("afterTargetExec", exception, tb)
                await dii.exec_lifecycle("afterExecution", exception, tb)

            gener = None
            if [inspect.isgenerator, isgenerator][cached_isinstance(result, Hashable)](
                result
            ):
                gener = result
                try:
                    result = next(gener)
                except StopIteration as e:
                    result = e.value
                else:
                    dii.alive_generator_dispatcher[-1].append((gener, False))
            elif [inspect.isasyncgen, isasyncgen][cached_isinstance(result, Hashable)](
                result
            ):
                gener = result
                try:
                    result = await gener.__anext__()
                except StopAsyncIteration:
                    return  # value = None
                else:
                    dii.alive_generator_dispatcher[-1].append((gener, True))

            if result.__class__ is Force:
                return result.content

            if result.__class__ is RemoveMe:
                if cached_isinstance(target, Listener):
                    if target in self.listeners:
                        self.listeners.pop(self.listeners.index(target))
            return result

    def postEvent(self, event: BaseEvent):
        self.loop.create_task(
            self.layered_scheduler(
                listener_generator=self.default_listener_generator(event.__class__),
                event=event,
            )
        )

    @staticmethod
    def event_class_generator(target=BaseEvent):
        for i in target.__subclasses__():
            yield i
            if i.__subclasses__():
                yield from Broadcast.event_class_generator(i)

    @staticmethod
    def findEvent(name: str):
        for i in Broadcast.event_class_generator():
            if i.__name__ == name:
                return i

    def getDefaultNamespace(self):
        return self.default_namespace

    def createNamespace(
        self, name, *, priority: int = 0, hide: bool = False, disabled: bool = False
    ):
        if self.containNamespace(name):
            raise ExistedNamespace(name, "has been created!")
        self.namespaces.append(
            Namespace(name=name, priority=priority, hide=hide, disabled=disabled)
        )
        return self.namespaces[-1]

    def removeNamespace(self, name):
        if self.containNamespace(name):
            for index, i in enumerate(self.namespaces):
                if i.name == name:
                    self.namespaces.pop(index)
                    return
        else:
            raise UnexistedNamespace(name)

    def containNamespace(self, name):
        for i in self.namespaces:
            if i.name == name:
                return True
        return False

    def getNamespace(self, name):
        if self.containNamespace(name):
            for i in self.namespaces:
                if i.name == name:
                    return i
        else:
            raise UnexistedNamespace(name)

    def hideNamespace(self, name):
        ns = self.getNamespace(name)
        ns.hide = True

    def unhideNamespace(self, name):
        ns = self.getNamespace(name)
        ns.hide = False

    def disableNamespace(self, name):
        ns = self.getNamespace(name)
        ns.disabled = True

    def enableNamespace(self, name):
        ns = self.getNamespace(name)
        ns.disabled = False

    def containListener(self, target):
        for i in self.listeners:
            if i.callable == target:
                return True
        return False

    def getListener(self, target):
        for i in self.listeners:
            if i.callable == target:
                return i

    def removeListener(self, target):
        self.listeners.remove(target)

    def includeZone(self, target_zone: Zone):
        if self.containNamespace(Zone.namespace.name):
            raise ValueError("the zone cannot import because its name has been used.")
        self.namespaces.append(target_zone.namespace)
        self.listeners.extend(target_zone.listeners)

    def receiver(
        self,
        event: Union[str, Type[BaseEvent]],
        priority: int = 16,
        dispatchers: List[Type[BaseDispatcher]] = [],
        namespace: Namespace = None,
        headless_decoraters: List[Decorater] = [],
        enable_internal_access: bool = False,
    ):
        if cached_isinstance(event, str):
            _name = event
            event = self.findEvent(event)
            if not event:
                raise InvaildEventName(_name, "is not vaild!")
        priority = (type(priority) == int) and priority or int(priority)  # 类型转换

        def receiver_wrapper(callable_target):
            may_listener = self.getListener(callable_target)
            if not may_listener:
                self.listeners.append(
                    Listener(
                        callable=callable_target,
                        namespace=namespace or self.getDefaultNamespace(),
                        inline_dispatchers=dispatchers,
                        priority=priority,
                        listening_events=[event],
                        headless_decoraters=headless_decoraters,
                        enable_internal_access=enable_internal_access,
                    )
                )
            else:
                if event not in may_listener.listening_events:
                    may_listener.listening_events.append(event)
                else:
                    raise RegisteredEventListener(
                        event.__name__, "has been registered!"
                    )
            return callable_target

        return receiver_wrapper
